@@ -1,5 +1,7 @@
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE PatternGuards #-}
 -- | DOM-based parsing and rendering.
 --
 -- This module requires that all entities be resolved at parsing. If you need
@@ -55,6 +57,9 @@ module Text.XML
       -- ** Rendering
     , R.RenderSettings
     , R.rsPretty
+    , R.rsNamespaces
+    , R.rsAttrOrder
+    , R.orderAttrs
       -- * Conversion
     , toXMLDocument
     , fromXMLDocument
@@ -74,6 +79,7 @@ import Data.XML.Types
     , ExternalID (..)
     )
 import Data.Typeable (Typeable)
+import Data.Data (Data)
 import Data.Text (Text)
 import qualified Text.XML.Stream.Parse as P
 import qualified Text.XML.Unresolved as D
@@ -88,39 +94,49 @@ import Data.ByteString (ByteString)
 import qualified Data.ByteString.Lazy as L
 import Control.Monad.ST (runST)
 import qualified Data.Set as Set
+import qualified Data.Map as Map
 import Data.Set (Set)
 
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Encoding as TLE
-import qualified Data.Conduit as C
+import Data.Conduit hiding (Source, Sink, Conduit)
 import qualified Data.Conduit.List as CL
 import qualified Data.Conduit.Binary as CB
 import System.IO.Unsafe (unsafePerformIO)
 import Control.Exception (throw)
-import Control.Monad.Trans.Resource (MonadUnsafeIO, runExceptionT)
+import Control.Monad.Trans.Resource (runExceptionT)
 import Control.Monad.Trans.Class (lift)
 import Data.Conduit.Lazy (lazyConsume)
 
+import qualified Text.Blaze as B
+import qualified Text.Blaze.Html as B
+import qualified Text.Blaze.Html5 as B5
+import qualified Text.Blaze.Internal as BI
+import Data.Monoid (mempty, mappend)
+import Data.String (fromString)
+import Data.List (foldl')
+import Control.Arrow (first)
+
 data Document = Document
     { documentPrologue :: Prologue
     , documentRoot :: Element
     , documentEpilogue :: [Miscellaneous]
     }
-  deriving (Show, Eq, Typeable)
+  deriving (Show, Eq, Typeable, Data)
 
 data Node
     = NodeElement Element
     | NodeInstruction Instruction
     | NodeContent Text
     | NodeComment Text
-  deriving (Show, Eq, Ord, Typeable)
+  deriving (Show, Eq, Ord, Typeable, Data)
 
 data Element = Element
     { elementName :: Name
-    , elementAttributes :: [(Name, Text)]
+    , elementAttributes :: Map.Map Name Text
     , elementNodes :: [Node]
     }
-  deriving (Show, Eq, Ord, Typeable)
+  deriving (Show, Eq, Ord, Typeable, Data)
 
 {-
 readFile :: FilePath -> ParseSettings -> IO (Either SomeException Document)
@@ -128,20 +144,29 @@ readFile_ :: FIlePath -> ParseSettings -> IO Document
 -}
 
 toXMLDocument :: Document -> X.Document
-toXMLDocument (Document a b c) = X.Document a (toXMLElement b) c
+toXMLDocument = toXMLDocument' def
+
+toXMLDocument' :: R.RenderSettings -> Document -> X.Document
+toXMLDocument' rs (Document a b c) = X.Document a (toXMLElement' rs b) c
 
 toXMLElement :: Element -> X.Element
-toXMLElement (Element name as nodes) =
+toXMLElement = toXMLElement' def
+
+toXMLElement' :: R.RenderSettings -> Element -> X.Element
+toXMLElement' rs (Element name as nodes) =
     X.Element name as' nodes'
   where
-    as' = map (\(x, y) -> (x, [X.ContentText y])) as
-    nodes' = map toXMLNode nodes
+    as' = map (\(x, y) -> (x, [X.ContentText y])) $ R.rsAttrOrder rs name as
+    nodes' = map (toXMLNode' rs) nodes
 
 toXMLNode :: Node -> X.Node
-toXMLNode (NodeElement e) = X.NodeElement $ toXMLElement e
-toXMLNode (NodeContent t) = X.NodeContent $ X.ContentText t
-toXMLNode (NodeComment c) = X.NodeComment c
-toXMLNode (NodeInstruction i) = X.NodeInstruction i
+toXMLNode = toXMLNode' def
+
+toXMLNode' :: R.RenderSettings -> Node -> X.Node
+toXMLNode' rs (NodeElement e) = X.NodeElement $ toXMLElement' rs e
+toXMLNode' _ (NodeContent t) = X.NodeContent $ X.ContentText t
+toXMLNode' _ (NodeComment c) = X.NodeComment c
+toXMLNode' _ (NodeInstruction i) = X.NodeInstruction i
 
 fromXMLDocument :: X.Document -> Either (Set Text) Document
 fromXMLDocument (X.Document a b c) =
@@ -160,7 +185,8 @@ fromXMLElement (X.Element name as nodes) =
     enodes = map fromXMLNode nodes
     (lnodes, rnodes) = partitionEithers enodes
     eas = map go as
-    (las, ras) = partitionEithers eas
+    (las, ras') = partitionEithers eas
+    ras = Map.fromList ras'
     go (x, y) =
         case go' [] id y of
             Left es -> Left es
@@ -181,7 +207,7 @@ fromXMLNode (X.NodeInstruction i) = Right $ NodeInstruction i
 readFile :: ParseSettings -> FilePath -> IO Document
 readFile ps fp = handle
     (throwIO . InvalidXMLFile fp)
-    (C.runResourceT $ CB.sourceFile (encodeString fp) C.$$ sinkDoc ps)
+    (runResourceT $ CB.sourceFile (encodeString fp) $$ sinkDoc ps)
 
 data XMLException = InvalidXMLFile FilePath SomeException
     deriving Typeable
@@ -199,45 +225,45 @@ parseLBS :: ParseSettings -> L.ByteString -> Either SomeException Document
 parseLBS ps lbs = runST
                 $ runExceptionT
                 $ CL.sourceList (L.toChunks lbs)
-           C.$$ sinkDoc ps
+           $$ sinkDoc ps
 
 parseLBS_ :: ParseSettings -> L.ByteString -> Document
 parseLBS_ ps = either throw id . parseLBS ps
 
-sinkDoc :: C.MonadThrow m
+sinkDoc :: MonadThrow m
         => ParseSettings
-        -> C.Sink ByteString m Document
-sinkDoc ps = P.parseBytes ps C.=$ fromEvents
+        -> Pipe l ByteString o u m Document
+sinkDoc ps = P.parseBytesPos ps >+> fromEvents
 
 parseText :: ParseSettings -> TL.Text -> Either SomeException Document
 parseText ps tl = runST
                 $ runExceptionT
                 $ CL.sourceList (TL.toChunks tl)
-           C.$$ sinkTextDoc ps
+           $$ sinkTextDoc ps
 
 parseText_ :: ParseSettings -> TL.Text -> Document
 parseText_ ps = either throw id . parseText ps
 
-sinkTextDoc :: C.MonadThrow m
+sinkTextDoc :: MonadThrow m
             => ParseSettings
-            -> C.Sink Text m Document
-sinkTextDoc ps = P.parseText ps C.=$ fromEvents
+            -> Pipe l Text o u m Document
+sinkTextDoc ps = P.parseText ps >+> fromEvents
 
-fromEvents :: C.MonadThrow m => C.Sink X.Event m Document
+fromEvents :: MonadThrow m => Pipe l P.EventPos o u m Document
 fromEvents = do
     d <- D.fromEvents
-    either (lift . C.monadThrow . UnresolvedEntityException) return $ fromXMLDocument d
+    either (lift . monadThrow . UnresolvedEntityException) return $ fromXMLDocument d
 
 data UnresolvedEntityException = UnresolvedEntityException (Set Text)
     deriving (Show, Typeable)
 instance Exception UnresolvedEntityException
 
-renderBytes :: MonadUnsafeIO m => R.RenderSettings -> Document -> C.Source m ByteString
-renderBytes rs doc = D.renderBytes rs $ toXMLDocument doc
+renderBytes :: MonadUnsafeIO m => R.RenderSettings -> Document -> Pipe l i ByteString u m ()
+renderBytes rs doc = D.renderBytes rs $ toXMLDocument' rs doc
 
 writeFile :: R.RenderSettings -> FilePath -> Document -> IO ()
 writeFile rs fp doc =
-    C.runResourceT $ renderBytes rs doc C.$$ CB.sinkFile (encodeString fp)
+    runResourceT $ renderBytes rs doc $$ CB.sinkFile (encodeString fp)
 
 renderLBS :: R.RenderSettings -> Document -> L.ByteString
 renderLBS rs doc =
@@ -250,3 +276,53 @@ renderLBS rs doc =
 
 renderText :: R.RenderSettings -> Document -> TL.Text
 renderText rs = TLE.decodeUtf8 . renderLBS rs
+
+instance B.ToMarkup Document where
+    toMarkup (Document _ root _) = B5.docType >> B.toMarkup root
+
+-- | Note that the special element name
+-- @{http://www.snoyman.com/xml2html}ie-cond@ with the single attribute @cond@
+-- is used to indicate an IE conditional comment.
+instance B.ToMarkup Element where
+    toMarkup (Element "{http://www.snoyman.com/xml2html}ie-cond" attrs children)
+      | [("cond", cond)] <- Map.toList attrs =
+        B.preEscapedToMarkup ("<!--[if " :: T.Text)
+        `mappend` B.preEscapedToMarkup cond
+        `mappend` B.preEscapedToMarkup ("]>" :: T.Text)
+        `mappend` mapM_ B.toMarkup children
+        `mappend` B.preEscapedToMarkup ("<![endif]-->" :: T.Text)
+
+    toMarkup (Element name' attrs children) =
+        if isVoid
+            then foldl' (B.!) leaf attrs'
+            else foldl' (B.!) parent attrs' childrenHtml
+      where
+        childrenHtml :: B.Html
+        childrenHtml =
+            case (name `elem` ["style", "script"], children) of
+                (True, [NodeContent t]) -> B.preEscapedToMarkup t
+                _ -> mapM_ B.toMarkup children
+
+        isVoid = nameLocalName name' `Set.member` voidElems
+
+        parent :: B.Html -> B.Html
+        parent = BI.Parent tag open close
+        leaf :: B.Html
+        leaf = BI.Leaf tag open (fromString " />")
+
+        name = T.unpack $ nameLocalName name'
+        tag = fromString name
+        open = fromString $ '<' : name
+        close = fromString $ concat ["</", name, ">"]
+
+        attrs' :: [B.Attribute]
+        attrs' = map goAttr $ map (first nameLocalName) $ Map.toList attrs
+        goAttr (key, value) = B.customAttribute (B.textTag key) $ B.toValue value
+
+instance B.ToMarkup Node where
+    toMarkup (NodeElement e) = B.toMarkup e
+    toMarkup (NodeContent t) = B.toMarkup t
+    toMarkup _ = mempty
+
+voidElems :: Set.Set T.Text
+voidElems = Set.fromAscList $ T.words $ T.pack "area base br col command embed hr img input keygen link meta param source track wbr"
